# -*- coding: utf-8 -*-
"""Defining the main representation of a MOF."""
import os
from collections import defaultdict
from typing import Dict, List, Optional, Union

import networkx as nx
import numpy as np
from backports.cached_property import cached_property
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.core import Lattice, Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from structuregraph_helpers.create import VestaCutoffDictNN

from .descriptors.sbu_dimensionality import get_structure_graph_dimensionality
from .fragmentor import FragmentationResult, run_fragmentation
from .fragmentor.branching_points import get_branch_points
from .utils import IStructure, pickle_dump, write_cif

__all__ = ["MOF"]


class MOF:
    """Main representation for a MOF structure.

    This container holds a structure and its associated graph.
    It also provides some convenience methods for getting neighbors
    or results of the fragmentation.

    Internally, this code typically uses IStructure objects to
    avoid bugs due to the mutability of Structure objects
    (e.g. the fragmentation code performs operations on the structure
    and we want to be sure that there is no impact on the input).

    Examples:
        >>> from moffragmentor import MOF
        >>> mof = MOF(structure, structure_graph)
        >>> # equivalent is to read from a cif file
        >>> mof = MOF.from_cif(cif_file)
        >>> # visualize the structure
        >>> mof.show_structure()
        >>> # get the neighbors of a site
        >>> mof.get_neighbor_indices(0)
        >>> # perform fragmentation
        >>> fragments mof.fragment()
    """

    def __init__(self, structure: Structure, structure_graph: StructureGraph):
        """Initialize a MOF object.

        Args:
            structure (Structure): Pymatgen Structure object
            structure_graph (StructureGraph): Pymatgen StructureGraph object
        """
        self._structure = structure
        # checker = MOFChecker(structure)
        # if checker.has_atomic_overlaps:
        #     raise ValueError("Structure has atomic overlaps.")

        self._structure_graph = structure_graph
        self._bridges = None
        self._nx_graph = None
        nx.set_node_attributes(
            self._structure_graph.graph,
            name="idx",
            values=dict(zip(range(len(structure_graph)), range(len(structure_graph)))),
        )

    def _reset(self):
        """Reset all parameters that are computed at some point."""
        self._bridges = None
        self._nx_graph = None

    def dump(self, path) -> None:
        """Dump this object as pickle file"""
        pickle_dump(self, path)

    def __len__(self) -> str:
        """Length of the MOF. Equivalent to the number of sites."""
        return len(self.structure)

    @property
    def structure(self):
        return self._structure

    @property
    def structure_graph(self):
        return self._structure_graph

    @cached_property
    def dimensionality(self):
        return get_structure_graph_dimensionality(self.structure_graph)

    @property
    def lattice(self) -> Lattice:
        return self.structure.lattice

    @property
    def composition(self) -> str:
        return self.structure.composition.alphabetical_formula

    @property
    def cart_coords(self) -> np.ndarray:
        return self.structure.cart_coords

    @cached_property
    def frac_coords(self) -> np.ndarray:
        """Return fractional coordinates of the structure.

        We cache this call as pymatgen seems to re-compute this.

        Returns:
            np.ndarray: fractional coordinates of the structure
                in array of shape (n_sites, 3)
        """
        return self.structure.frac_coords

    @classmethod
    def from_cif(
        cls,
        cif: Union[str, os.PathLike],
        symprec: Optional[float] = None,
        angle_tolerance: Optional[float] = None,
        get_primitive: bool = True,
    ):
        """Initialize a MOF object from a cif file.

        Note that this method, by default, symmetrizes the structure.

        Args:
            cif (str): path to the cif file
            symprec (float, optional): Symmetry precision
            angle_tolerance (float, optional): Angle tolerance
            get_primitive (bool): Whether to get the primitive cell

        Returns:
            MOF: MOF object
        """
        # using the IStructure avoids bugs where somehow the structure changes
        structure = IStructure.from_file(cif)
        if get_primitive:
            structure = structure.get_primitive_structure(0.5)
        if symprec is not None and angle_tolerance is not None:
            spga = SpacegroupAnalyzer(structure, symprec=symprec, angle_tolerance=angle_tolerance)
            structure = spga.get_conventional_standard_structure()
        structure = IStructure.from_sites(structure)
        structure_graph = StructureGraph.with_local_env_strategy(structure, VestaCutoffDictNN)
        return cls(structure, structure_graph)

    @classmethod
    def from_structure(
        cls,
        structure: Structure,
        symprec: Optional[float] = 0.5,
        angle_tolerance: Optional[float] = 5,
    ):
        if (symprec is not None) and (angle_tolerance is not None):
            spga = SpacegroupAnalyzer(structure, symprec=symprec, angle_tolerance=angle_tolerance)
            structure = spga.get_conventional_standard_structure()
        structure = IStructure.from_sites(structure)
        structure_graph = StructureGraph.with_local_env_strategy(structure, VestaCutoffDictNN)
        return cls(structure, structure_graph)

    def _is_terminal(self, index):
        return len(self.get_neighbor_indices(index)) == 1

    @cached_property
    def terminal_indices(self) -> List[int]:
        """Return the indices of the terminal sites.

        A terminal site is a site that has only one neighbor.
        And is connected via a bridge to the rest of the structure.
        That means, splitting the bond between the terminal site
        and the rest of the structure will increase the number
        of connected components.

        Typical examples of terminal sites are hydrogren atoms,
        or halogen functional groups.

        Returns:
            List[int]: indices of the terminal sites
        """
        return [i for i in range(len(self.structure)) if self._is_terminal(i)]

    def _get_nx_graph(self):
        if self._nx_graph is None:
            self._nx_graph = nx.Graph(self.structure_graph.graph.to_undirected())
        return self._nx_graph

    def _leads_to_terminal(self, edge):
        sorted_edge = sorted(edge)
        try:
            bridge_edge = self.bridges[sorted_edge[0]]
            return sorted_edge[1] in bridge_edge
        except KeyError:
            return False

    @cached_property
    def nx_graph(self):
        """Structure graph as networkx graph object"""
        return self._get_nx_graph()

    def _generate_bridges(self) -> Dict[int, int]:
        if self._bridges is None:
            bridges = list(nx.bridges(self.nx_graph))

            bridges_dict = defaultdict(list)
            for key, value in bridges:
                bridges_dict[key].append(value)

            self._bridges = dict(bridges_dict)
        return self._bridges

    @cached_property
    def bridges(self) -> Dict[int, int]:
        """Get a dictionary of bridges.

        Bridges are edges in a graph that, if deleted, increase the number of connected components.

        Returns:
            Dict[int, int]: dictionary of bridges
        """
        return self._generate_bridges()

    @cached_property
    def adjaceny_matrix(self):
        return nx.adjacency_matrix(self.structure_graph.graph)

    def show_adjacency_matrix(self, highlight_metals=False):
        """Plot structure graph as adjaceny matrix"""
        import matplotlib.pylab as plt  # pylint:disable=import-outside-toplevel

        matrix = self.adjaceny_matrix.todense()
        if highlight_metals:
            cols = np.nonzero(matrix[self.metal_indices, :])
            rows = np.nonzero(matrix[:, self.metal_indices])
            matrix[self.metal_indices, cols] = 2
            matrix[rows, self.metal_indices] = 2
        plt.imshow(self.adjaceny_matrix.todense(), cmap="Greys_r")

    @cached_property
    def _branching_indices_list(self):
        return get_branch_points(self)

    def _is_branch_point(self, index):
        return index in self._branching_indices_list

    @cached_property
    def metal_indices(self) -> List[int]:
        return [i for i, species in enumerate(self.structure.species) if species.is_metal]

    @cached_property
    def h_indices(self) -> List[int]:
        return [i for i, species in enumerate(self.structure.species) if str(species) == "H"]

    @cached_property
    def c_indices(self) -> List[int]:
        return [i for i, species in enumerate(self.structure.species) if str(species) == "C"]

    def get_neighbor_indices(self, site: int) -> List[int]:
        """Get list of indices of neighboring sites."""
        return [site.index for site in self.structure_graph.get_connected_sites(site)]

    def get_symbol_of_site(self, site: int) -> str:
        """Get elemental symbol of site indexed site."""
        return str(self.structure[site].specie)

    def show_structure(self):
        """Visualize structure using nglview."""
        import nglview  # pylint:disable=import-outside-toplevel

        return nglview.show_pymatgen(self.structure)

    def _fragment(
        self, check_dimensionality, create_single_metal_bus, break_organic_nodes_at_metal
    ):
        fragmentation_result = run_fragmentation(
            self,
            check_dimensionality,
            create_single_metal_bus,
            break_organic_nodes_at_metal=break_organic_nodes_at_metal,
        )
        return fragmentation_result

    def fragment(
        self,
        check_dimensionality: bool = True,
        create_single_metal_bus: bool = False,
        break_organic_nodes_at_metal: bool = True,
    ) -> FragmentationResult:
        """Split the MOF into building blocks.

        The building blocks are linkers, nodes, bound,
        unbound solvent, net embedding of those building blocks.

        Args:
            check_dimensionality (bool): Check if the node is 0D.
                If not, split into isolated metals.
                Defaults to True.
            create_single_metal_bus (bool): Create a single metal BUs.
                Defaults to False.
            break_organic_nodes_at_metal (bool): Break nodes into single metal BU
                if they appear "too organic".

        Returns:
            FragmentationResult: FragmentationResult object.
        """
        return self._fragment(
            check_dimensionality,
            create_single_metal_bus,
            break_organic_nodes_at_metal=break_organic_nodes_at_metal,
        )

    def _get_cif_text(self) -> str:
        return write_cif(self.structure, self.structure_graph, [])

    def write_cif(self, filename) -> None:
        """Write the structure to a CIF file."""
        with open(filename, "w", encoding="utf8") as file_handle:
            file_handle.write(self._get_cif_text())



# -*- coding: utf-8 -*-
"""Testing functions related to molecule creation"""
from moffragmentor.molecule import NonSbuMolecule, NonSbuMoleculeCollection


def test_molecule_creation(get_methane_molecule_and_graph):
    """Make sure we can create a molecule object without any issues"""
    methane, methane_graph = get_methane_molecule_and_graph
    mol = NonSbuMolecule(methane, methane_graph, [0, 1, 2, 3])
    assert str(mol) == "C1 H4"
    assert len(mol) == 5


def test_molecule_creation_from_structure_graph(get_p_linker_with_floating):
    """Make sure we can create a molecule object from a structure graph"""
    mof = get_p_linker_with_floating
    graph = mof.structure_graph
    indices = [167, 159, 251]
    mol = NonSbuMolecule.from_structure_graph_and_indices(graph, indices)
    assert str(mol) == "H2 O1"


def test_get_molecule_collection(get_methane_molecule_and_graph):
    """Make sure we can create a collection of molecules"""
    methane, methane_graph = get_methane_molecule_and_graph
    mol = NonSbuMolecule(methane, methane_graph, [0, 1, 2, 3])
    molecule_collection = NonSbuMoleculeCollection([mol, mol])
    assert molecule_collection.composition == {"C1 H4": 2}

    # we should be able to sum the collections
    new_collection = molecule_collection + molecule_collection
    assert len(new_collection) == 4
    assert new_collection.composition == {"C1 H4": 4}

    # the old one is unmodified
    assert len(molecule_collection) == 2



# -*- coding: utf-8 -*-
# pylint:disable=no-member
"""Methods to rank molecules according to some measure of similarity."""
from rdkit import Chem, DataStructs
from rdkit.Chem import rdFMCS


def tanimoto_rank(smiles_reference, smiles, additional_attributes=None):
    """Rank SMILES based on the Tanimoto similarity to the reference smiles."""
    distances = []
    mol = Chem.MolFromSmiles(smiles_reference)
    fp = Chem.RDKFingerprint(mol)
    for i, smile in enumerate(smiles):
        mol2 = Chem.MolFromSmiles(smile)
        fp2 = Chem.RDKFingerprint(mol2)
        add_attr = additional_attributes[i] if additional_attributes else None
        distances.append((smile, DataStructs.FingerprintSimilarity(fp, fp2), add_attr))

    sorted_by_second = sorted(distances, key=lambda tup: tup[1])
    return sorted_by_second


def mcs_rank(
    smiles_reference, smiles, additional_attributes=None
):  # pylint:disable=too-many-locals
    """Rank SMILES based on the maximum common substructure to the reference smiles."""
    distances = []
    mol = Chem.MolFromSmiles(smiles_reference)
    num_atoms = mol.GetNumAtoms()
    num_bonds = mol.GetNumBonds()
    for i, smile in enumerate(smiles):
        mol2 = Chem.MolFromSmiles(smile)
        res = rdFMCS.FindMCS([mol, mol2], completeRingsOnly=True, ringMatchesRingOnly=True)
        smarts = res.smartsString
        bond_diff = num_bonds - res.numBonds
        atom_diff = num_atoms - res.numAtoms
        add_attr = additional_attributes[i] if additional_attributes else None
        distances.append((smile, smarts, bond_diff, atom_diff, add_attr))

    sorted_by_atom_diff = sorted(distances, key=lambda tup: tup[-2])
    return sorted_by_atom_diff



# -*- coding: utf-8 -*-
"""Classes for the basic building blocks"""
from .linker import Linker  # noqa:F401
from .linkercollection import LinkerCollection  # noqa:F401
from .node import Node  # noqa:F401
from .nodecollection import NodeCollection  # noqa:F401
from .sbu import SBU  # noqa:F401



# -*- coding: utf-8 -*-
"""Dealing with molecules (SBUs) and solvents"""
from .nonsbumolecule import NonSbuMolecule  # noqa: F401
from .nonsbumoleculecollection import NonSbuMoleculeCollection  # noqa: F401



# -*- coding: utf-8 -*-
"""Command line interfaces"""
import click

from ..harvester.harvest_building_blocks import harvest_directory


@click.command("cli")
@click.argument("indir", type=click.Path(exists=True))
@click.argument("dumpdir", default=None)
@click.option("--njobs", "-n", default=1, type=int)
@click.option("--reverse", is_flag=True)
@click.option("--offset", "-o", default=0, type=int)
def run_harvest(indir, dumpdir, njobs, reverse, offset):
    harvest_directory(indir, njobs, dumpdir, reverse=reverse, offset=offset)



import pytest

from moffragmentor.utils import get_molecule_mass


def test_get_molecule_mass(get_methane_molecule_and_graph):
    mol, graph = get_methane_molecule_and_graph

    assert get_molecule_mass(mol) == pytest.approx(16.04303, 0.1)



# -*- coding: utf-8 -*-
"""Functionality to fragment all MOFs in a folder of CIF files"""
# pylint:disable=invalid-name
import concurrent.futures
import os
import pickle
from functools import partial
from glob import glob
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd
import timeout_decorator
from loguru import logger

from ..descriptors.sbu_dimensionality import get_sbu_dimensionality
from ..mof import MOF
from ..utils import get_linker_connectivity, make_if_not_exists


def load_failed():
    try:
        with open("failed_cifs.pkl", "rb") as handle:
            failed_cifs = pickle.load(handle)
            return failed_cifs
    except Exception:  # pylint:disable=broad-except
        return set()


def sbu_descriptors(  # pylint:disable=too-many-arguments
    sbu, mof, bb_type="linker", topology="", dimensionality=np.nan, connectivity=np.nan
):
    descriptors = sbu.descriptors
    descriptors["smiles"] = sbu.smiles
    descriptors["type"] = bb_type
    descriptors["topology"] = topology
    descriptors["mof_dimensionality"] = dimensionality
    descriptors["coordination"] = sbu.coordination
    descriptors["connectivity"] = connectivity
    descriptors["dimensionality"] = get_sbu_dimensionality(
        mof, sbu._original_indices  # pylint:disable=protected-access
    )
    return {**descriptors}


MAX_ATOMS = 3000


class Harvester:
    """Orchestrating the large-scale mining of building blocks."""

    def __init__(self, mof: MOF, outdir: Optional[Union[str, os.PathLike]] = None):
        """Initialize the harvester.

        Args:
            mof (MOF): MOF to be harvested.
            outdir (Optional[Union[str, os.PathLike]], optional):
                Name of/path to output directory. Defaults to None.
        """
        self.mof = mof
        self.outdir = outdir

    @classmethod
    def from_cif(cls, cif, outdir=None):
        mof = MOF.from_cif(cif)
        if len(mof.structure) > MAX_ATOMS:
            raise ValueError("Structure too large")
        return cls(mof, outdir)

    def run_harvest(self):
        if self.outdir is not None:
            self.mof.dump(os.path.join(self.outdir, "mof.pkl"))
        descriptors = []

        parts = self.mof.fragment()
        try:
            topology = parts.net_embedding.rcsr_code
        except Exception:  # pylint:disable=broad-except
            topology = ""
        try:
            dimensionality = self.mof.dimensionality
        except Exception:  # pylint:disable=broad-except
            dimensionality = np.nan
        # ToDo: Fixme for the new API
        edge_dict = parts.net_embedding._edge_dict
        linker_connectivity = get_linker_connectivity(edge_dict)
        for i, linker in enumerate(parts.linkers):
            if self.outdir is not None:
                linker.dump(os.path.join(self.outdir, f"linker_{i}.pkl"))
            descriptors.append(
                sbu_descriptors(
                    linker,
                    self.mof,
                    bb_type="linker",
                    topology=topology,
                    dimensionality=dimensionality,
                    connectivity=linker_connectivity[i],
                )
            )
        for i, node in enumerate(parts.nodes):
            if self.outdir is not None:
                node.dump(os.path.join(self.outdir, f"node_{i}.pkl"))
            descriptors.append(
                sbu_descriptors(
                    node,
                    self.mof,
                    bb_type="node",
                    topology=topology,
                    dimensionality=dimensionality,
                    connectivity=len(edge_dict[i]),
                )
            )

        df = pd.DataFrame(descriptors)
        if self.outdir is not None:
            df.to_csv(os.path.join(self.outdir, "descriptors.csv"), index=False)
        return df


def harvest_w_timeout(cif, dumpdir=None):
    try:
        return harvest_cif(cif, dumpdir=dumpdir)
    except Exception as e:  # pylint:disable=broad-except
        logger.exception("%s. Exception: %s.", cif, e)
        return None


@timeout_decorator.timeout(200)
def harvest_cif(cif, dumpdir=None):
    try:
        stem = Path(cif).stem
        if dumpdir is not None:
            path = os.path.join(dumpdir, stem)
            make_if_not_exists(path)
            dumpdir = path
        harvester = Harvester.from_cif(cif, dumpdir)
        if len(harvester.mof) > 500:
            raise ValueError("Structure too large")
        df = harvester.run_harvest()
        return df
    except Exception as execpt:  # pylint:disable=broad-except
        logger.exception("%s. Exception: %s.", cif, execpt)
        return None


def harvest_directory(  # pylint:disable=too-many-branches, too-many-arguments, too-many-locals
    directory,
    njobs=1,
    outdir=None,
    skip_existing=True,
    reverse=False,
    offset: int = 0,
    qmof: bool = False,
):
    all_cifs = glob(os.path.join(directory, "*.cif"))
    if outdir is not None:
        make_if_not_exists(outdir)

    harvest_partial = partial(harvest_w_timeout, dumpdir=outdir)

    if skip_existing:
        existing_stems = {
            str(Path(p).parents[0]).split("_", maxsplit=1)[0]
            for p in glob(os.path.join(outdir, "*", "descriptors.csv"))
        }
        existing_stems.update(load_failed())
        filtered_all_cifs = []
        for cif in all_cifs:
            if str(Path(cif).stem).split("_", maxsplit=1)[0] not in existing_stems:
                if qmof:
                    if "FSR" in cif:
                        filtered_all_cifs.append(cif)
                else:
                    filtered_all_cifs.append(cif)

    else:
        filtered_all_cifs = all_cifs

    filtered_all_cifs = sorted(filtered_all_cifs, reverse=reverse)
    all_res = []
    with concurrent.futures.ProcessPoolExecutor(max_workers=njobs) as executor:
        for i, res in enumerate(executor.map(harvest_partial, filtered_all_cifs[offset:])):
            try:
                if res is not None:
                    all_res.append(res)
                else:
                    logger.error("Failed for %s", filtered_all_cifs[i])
            except concurrent.futures.process.BrokenProcessPool as ex:
                logger.error("Broken process pool due to %s", ex)

    df = pd.concat(all_res)
    if outdir is None:
        df.to_csv("harvest_results.csv", index=False)

    else:
        df.to_csv(os.path.join(outdir, "harvest_results.csv"), index=False)



# -*- coding: utf-8 -*-
"""Based on the node location, locate the linkers"""
from typing import Tuple

import numpy as np
from pymatgen.core import Structure
from structuregraph_helpers.subgraph import get_subgraphs_as_molecules

from .molfromgraph import wrap_molecule
from ..sbu import Linker, LinkerCollection
from ..utils import _flatten_list_of_sets

__all__ = ["create_linker_collection", "identify_linker_binding_indices"]


def _is_site_in_cell(coord):
    return np.all((coord >= 1) & (coord <= 2))


def _any_site_in_cell(coords):
    return np.any([_is_site_in_cell(coord) for coord in coords])


def _pick_central_linker_indices(mof, coords):
    selected = []
    coordinates = []
    for i, coord in enumerate(coords):
        # we mimic that the central (1,1,1)-(2,2,2) unit cell is the
        # original unit cell
        frac = mof.structure.lattice.get_fractional_coords(coord)
        if _any_site_in_cell(frac):
            # the (1,1,1) coordinate is the origin of the original unit cell
            trans_frac = frac - np.array([1, 1, 1])
            old_cart = mof.structure.lattice.get_cartesian_coords(trans_frac)
            selected.append(i)
            coordinates.append(old_cart)
    return selected, coordinates


def number_branching_points_in_cell(coordinates, lattice):
    in_cell = 0
    for coord in coordinates:
        frac_coord = lattice.get_fractional_coords(coord)
        if np.all(frac_coord <= 1) & np.all(frac_coord >= 0):
            in_cell += 1

    return in_cell


def _create_linkers_from_node_location_result(  # pylint:disable=too-many-locals
    mof, node_location_result, node_collection, unbound_solvent, bound_solvent
) -> Tuple[LinkerCollection, dict]:
    linkers = []

    # First step: remove everything that is node or unbound solvent
    # bound solvent is part of the node by default
    all_node_indices = set()
    all_persistent_non_metal_bridges = set()
    for node_indices in node_location_result.nodes:
        all_node_indices.update(node_indices)

    for node in node_collection:
        all_persistent_non_metal_bridges.update(
            _flatten_list_of_sets(
                node._persistent_non_metal_bridged  # pylint:disable=protected-access
            )
        )

    not_linker_indices = (
        (
            all_node_indices
            & (
                node_location_result.connecting_paths
                - node_location_result.branching_indices
                - node_location_result.binding_indices
                - set(sum(node_location_result.to_terminal_from_branching.values(), []))
            )
        )
        | set(unbound_solvent.indices)
        | set(bound_solvent.indices)
        | set(mof.metal_indices) & all_node_indices
        # some metals might also be in the linker, e.g., in porphyrins
    )

    # potential_linker_indices = set(list(range(len(mof.structure)))) - not_linker_indices
    # get terminal indices we need to keep in the linker

    # terminal_indices = []
    # for linker_index in potential_linker_indices:
    #     for neighbor in mof.get_neighbor_indices(linker_index):
    #         if mof._is_terminal(neighbor):
    #             terminal_indices.append(neighbor)

    graph_ = mof.structure_graph.__copy__()
    graph_.structure = Structure.from_sites(graph_.structure.sites)
    graph_.remove_nodes(not_linker_indices)

    # Second step: extract the connected components
    # return all as molecules
    mols, graphs, idxs, centers, coordinates = get_subgraphs_as_molecules(
        graph_,
        return_unique=False,
        filter_in_cell=False,
        disable_boundary_crossing_check=True,
        prune_long_edges=True,
    )

    # Third: pick those molecules that are closest to the UC
    # ToDo: we should be able to skip this
    linker_indices, coords = _pick_central_linker_indices(mof, coordinates)
    found_frac_centers = set()
    found_hashes = set()
    for i, linker_index in enumerate(linker_indices):
        idx = idxs[linker_index]
        center = centers[linker_index]
        coords_ = coords[i]
        branching_indices = node_location_result.branching_indices & set(idx)
        mol, mapping = wrap_molecule(idx, mof)
        linker = Linker(
            molecule=mol,
            molecule_graph=graphs[linker_index],
            center=center,
            graph_branching_indices=branching_indices,
            closest_branching_index_in_molecule=branching_indices,
            binding_indices=identify_linker_binding_indices(
                mof,
                node_location_result.connecting_paths,
                idx,
            ),
            coordinates=coords_,
            original_indices=idx,
            connecting_paths=[],
            molecule_original_indices_mapping=mapping,
        )
        frac_center = mof.structure.lattice.get_fractional_coords(mol.center_of_mass)
        frac_center -= np.floor(frac_center)
        frac_center = (
            np.round(frac_center[0], 4),
            np.round(frac_center[1], 4),
            np.round(frac_center[2], 4),
        )
        if linker.hash not in found_hashes and frac_center not in found_frac_centers:
            found_hashes.add(linker.hash)
            linkers.append(linker)
            found_frac_centers.add(frac_center)
    linker_collection = LinkerCollection(linkers)

    return linker_collection


def create_linker_collection(
    mof,
    node_location_result,
    node_collection,
    unbound_solvents,
    bound_solvents,
) -> Tuple[LinkerCollection, dict]:
    """Based on MOF, node locaion and unbound solvent location locate the linkers"""
    linker_collection = _create_linkers_from_node_location_result(
        mof, node_location_result, node_collection, unbound_solvents, bound_solvents
    )
    return linker_collection


def identify_linker_binding_indices(mof, connecting_paths, indices):
    relevant_indices = connecting_paths & set(indices)

    filtered_indices = []

    metal_indices = set(mof.metal_indices)
    for idx in relevant_indices:
        neighbors = mof.get_neighbor_indices(idx)
        if not metal_indices.isdisjoint(set(neighbors)):
            filtered_indices.append(idx)

    return filtered_indices


def _check_linker(linker, mof):
    # check that no linker has more than two metals
    metal_indices = set(mof.metal_indices)
    return len(metal_indices & set(linker.original_indices)) < 2


def check_linkers(linker_collection, mof):
    """Check if the linkers are valid"""
    return all(_check_linker(linker, mof) for linker in linker_collection)



# -*- coding: utf-8 -*-
"""Collections of molecules, e.g. bound solvents and non-bound solvents."""
from collections import Counter
from typing import List

from .nonsbumolecule import NonSbuMolecule


class NonSbuMoleculeCollection:
    """Class to handle collections of molecules.

    For example, bound solvents and non-bound solvents.
    """

    def __init__(self, non_sbu_molecules: List[NonSbuMolecule]):
        """Construct a NonSbuMoleculeCollection.

        Args:
            non_sbu_molecules (List[NonSbuMolecule]): List of NonSbuMolecule objects.
        """
        self.molecules = non_sbu_molecules
        self._composition = None
        # currently also contains indices from the supercell expansion
        self.indices = sum([molecule.indices for molecule in self.molecules], [])

    def __len__(self):
        """Return number of molecules in the collection."""
        return len(self.molecules)

    def __repr__(self):
        """Return a string representation of the collection."""
        return f"NonSbuMoleculeCollection({self._get_composition()})"

    def __getitem__(self, index):
        """Get a molecule from the collection."""
        return self.molecules[index]

    def __next__(self):
        """Iterate over the molecules in the collection."""
        for molecule in self.molecules:
            yield molecule

    def __add__(self, other: "NonSbuMoleculeCollection"):  # noqa: F821
        """Add two collections of molecules."""
        molecules = self.molecules + other.molecules
        return NonSbuMoleculeCollection(molecules)

    def _get_composition(self):
        if self._composition is None:
            composition = []
            for mol in self.molecules:
                composition.append(str(mol.composition))
            composition_counter = Counter(composition)
            self._composition = dict(composition_counter)

        return self._composition

    @property
    def composition(self) -> str:
        """Get a string describing the composition."""
        return self._get_composition()



# -*- coding: utf-8 -*-
"""Describing a collection of nodes"""
from pymatgen.core import Molecule

from .sbucollection import SBUCollection

__all__ = ["NodeCollection"]


class NodeCollection(SBUCollection):
    """Collection of node building blocks"""

    @property
    def building_block_composition(self):
        if self._composition is None:
            self._composition = ["".join(["N", i]) for i in self.sbu_types]
        return self._composition

    def _node_structure(self):
        all_node_sites = []

        for n in self:
            n.molecule.coords = n.cart_coords
            s = n.molecule.sites
            all_node_sites.extend(s)

        super_node = Molecule.from_sites(all_node_sites)
        return super_node

    def show_node_structure(self):
        import nglview as nv

        super_node = self._node_structure()
        return nv.show_pymatgen(super_node)



# -*- coding: utf-8 -*-
"""Testing some attributes of the MOF class"""
import os

import pytest
from pymatgen.core import Structure

THIS_DIR = os.path.dirname(os.path.realpath(__file__))


def check_mof_creation_from_cif(get_cuiiibtc_mof):
    mof = get_cuiiibtc_mof
    assert isinstance(mof, object)
    assert isinstance(mof.structure, Structure)


def test_is_terminal(get_cuiiibtc_mof):
    """An atom is terminal when it only has one neighbor"""
    mof = get_cuiiibtc_mof
    assert mof._is_terminal(32)  # pylint:disable=protected-access
    assert not mof._is_terminal(129)  # pylint:disable=protected-access
    assert not mof._is_terminal(176)  # pylint:disable=protected-access


def test__leads_to_terminal(get_cuiiibtc_mof):
    mof = get_cuiiibtc_mof
    assert mof._leads_to_terminal((15, 183))  # pylint:disable=protected-access


@pytest.mark.slow
def test_fragmentation(get_cuiiibtc_mof):
    mof = get_cuiiibtc_mof
    fragments = mof.fragment()
    # topocryst.com fails here
    assert fragments.net_embedding.rcsr_code == "mmm"


@pytest.mark.slow
def test_fragmentation_single_metal(get_single_metal_mof, get_single_metal_mof_previously_failing):
    mof = get_single_metal_mof
    fragments = mof.fragment()
    assert len(fragments.nodes) == 2
    assert fragments.linkers[0].is_edge
    # this seems a bit flaky, but in the past: fragments.net_embedding.rcsr_code == "dia"

    mof = get_single_metal_mof_previously_failing
    fragments = mof.fragment()
    assert len(fragments.linkers) == 6


@pytest.mark.slow
def test_fragmentation_ag_n_compound(get_agn_mof):
    mof = get_agn_mof
    fragments = mof.fragment()
    pubchem_res = fragments.linkers[0].search_pubchem()
    assert pubchem_res[0][0].cid == 59188427

    # actually not sure if this test was correct
    # topos also just gives "unknown topology" for all clustering techniques
    # as does mofid
    # in particular, bex is also a 2D periodic compound
    # assert fragments.net_embedding.rcsr_code == "bex"


def _get_distances(cs):
    distances = []
    for i, site in enumerate(cs):
        for j, site2 in enumerate(cs):
            if i != j:
                distances.append(site.distance(site2))
    return distances


@pytest.mark.slow
def test_hkust(get_hkust_mof):
    mof = get_hkust_mof
    fragments = mof.fragment()
    assert fragments.net_embedding.rcsr_code == "tbo"
    for linker in fragments.linkers:
        cs = linker._get_connected_sites_structure()
        distances = _get_distances(cs)
        assert max(distances) < 8

    for node in fragments.nodes:
        cs = node._get_connected_sites_structure()
        distances = _get_distances(cs)
        assert max(distances) < 8


@pytest.mark.slow
def test_mof5(get_mof5):
    mof = get_mof5
    fragments = mof.fragment()
    assert fragments.net_embedding.rcsr_code == "pcu"



# -*- coding: utf-8 -*-
"""Test the filter modules"""
from moffragmentor.fragmentor.filter import _filter_branch_points


def test__filter_branching_points(get_dicarboxy_biphenyl_graph):
    """Branching points are only those for
    which the path to a node does not cross another branching point"""
    g, metal, branching = get_dicarboxy_biphenyl_graph
    filtered_indices = _filter_branch_points(branching, metal, g)
    assert len(filtered_indices) == 2
    assert "n2" in filtered_indices
    assert "n15" in filtered_indices



# -*- coding: utf-8 -*-
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))


# -- Project information -----------------------------------------------------

project = "moffragmentor"
copyright = "2022, Kevin Maik Jablonka"
author = "Kevin Maik Jablonka"

# The full version, including alpha/beta/rc tags
release = "0.0.2"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.mathjax",
    "sphinx_copybutton",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosectionlabel",
]

copybutton_selector = "div:not(.no-copy)>div.highlight pre"
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


numfig = True


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
if os.path.exists("_static/logo.png"):
    html_logo = "_static/logo.png"

# Register the theme as an extension to generate a sitemap.xml


html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/kjappelbaum/moffragmentor",
    "repo_url": "https://github.com/kjappelbaum/moffragmentor",
    "repo_name": "moffragmentor",
    "repo_type": "github",
    # "source_repository": "https://github.com/pradyunsg/furo/",
    "source_branch": "main",
    "source_directory": "docs/",
    "edit_uri": "https://github.com/kjappelbaum/moffragmentor/docs/source",
}


html_show_sphinx = False


# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = True



# -*- coding: utf-8 -*-
"""Methods on structure graphs"""


from pymatgen.core import Structure

from . import _not_relevant_structure_indices


def is_periodic(mof, indices):
    graph_ = mof.structure_graph.__copy__()
    graph_.structure = Structure.from_sites(graph_.structure.sites)
    to_delete = _not_relevant_structure_indices(mof.structure, indices)
    graph_.remove_nodes(to_delete)
    mols = graph_.get_subgraphs_as_molecules()
    return len(mols) > 0



# -*- coding: utf-8 -*-
"""Make sure we do not have a regression
in determining the RCSR code"""

import os

import pytest

from moffragmentor import MOF

THIS_DIR = os.path.dirname(os.path.realpath(__file__))
TEST_FILE_DIR = os.path.join(THIS_DIR, "..", "test_files")


def get_rscr_code(cif):
    path = os.path.join(TEST_FILE_DIR, cif)
    mof = MOF.from_cif(path)
    parts = mof.fragment()
    return parts.net_embedding.rcsr_code


@pytest.mark.slow
def test_net_detection():
    """On a set of structures from the review and beyond.
    Not that we only give the parent net, not potential subnets.
    """
    assert get_rscr_code("mof-5_cellopt.cif") == "pcu"
    assert get_rscr_code("HKUST-1.cif") == "tbo"
    assert get_rscr_code("QEFWUV_clean.cif") == "csq"
    assert get_rscr_code("LASYOU.cif") == "nbo"
    assert get_rscr_code("OFOCUI_clean.cif") == "nbo"
    assert get_rscr_code("WETPES_clean.cif") == "ssa"
    assert get_rscr_code("RUSWOT_clean.cif") == "pcu"



# -*- coding: utf-8 -*-
"""Test ranking of molecules according to similarity"""
from moffragmentor.utils.mol_compare import mcs_rank, tanimoto_rank

reference_smiles = "C1=CC=CC=C1"

query_smiles = ["C1=CC=CC=N1", "C1=CC=CC=P1", "CC1=CC=CC=C1"]


def test_tanimoto_rank():
    ranked_smiles = tanimoto_rank(reference_smiles, query_smiles)
    assert ranked_smiles[0][0] == "C1=CC=CC=N1"
    assert ranked_smiles[1][0] == "C1=CC=CC=P1"
    assert ranked_smiles[2][0] == "CC1=CC=CC=C1"


def test_mcs_rank():
    ranked_smiles = mcs_rank(reference_smiles, query_smiles)
    assert ranked_smiles[1][0] == "C1=CC=CC=N1"
    assert ranked_smiles[2][0] == "C1=CC=CC=P1"
    assert ranked_smiles[0][0] == "CC1=CC=CC=C1"



"""
Routines for finding branching points in a structure graph of a MOF.

Note that those routines do not work for other reticular materials
as they assume the presence of a metal.
"""
from operator import itemgetter
from typing import List, Tuple

import networkx as nx
import numpy as np
from loguru import logger
from more_itertools import pairwise
from pymatgen.core import Structure
from structuregraph_helpers.subgraph import get_subgraphs_as_molecules

from moffragmentor.utils import _not_relevant_structure_indices

from .. import mof

__all__ = ("get_branch_points",)


def get_distances_to_metal(
    mof: "mof.MOF", site: int
) -> List[float]:  # noqa: F821 - forward reference
    """For a given site, return the distances to all metals in the MOF."""
    distances = []
    for i in mof.metal_indices:
        distances.append(mof.structure.get_distance(site, i))
    return distances


def has_bridge_in_path(mof: "mof.MOF", path: List[int]) -> bool:
    return any(mof._leads_to_terminal(edge) for edge in pairwise(path))


def get_two_edge_paths_from_site(mof: "mof.MOF", site: int) -> List[List[int]]:
    """
    Return all two edge paths from a site.

    Args:
        mof (MOF): MOF object
        site (int): index of site that is to be probed.

    Returns:
        List[List[int]]: List of all two edge paths from the site.

    Example:
        >>> mof = MOF(...)
        >>> get_two_edge_paths_from_site(mof, 0)
        [[0, 1, 2], [0, 3, 4], [0, 5, 6]]
    """
    paths = []
    for i in mof.get_neighbor_indices(site):
        for j in mof.get_neighbor_indices(i):
            if j != site:
                paths.append([site, i, j])

    return paths


def has_metal_in_path(mof: "MOF", path: List[int]) -> bool:  # noqa: F821 - forward reference
    """Return True if the path contains a metal."""
    return any(site in mof.metal_indices for site in path)


def has_non_bridge_path_with_metal(mof: "mof.MOF", site: int) -> bool:
    """Return True if the MOF has a non-bridge path with a meta.l"""
    return any(
        (not has_bridge_in_path(mof, path)) and (has_metal_in_path(mof, path))
        for path in get_two_edge_paths_from_site(mof, site)
    )


def _is_branch_point(mof: "mof.MOF", index: int, allow_metal: bool = False) -> bool:
    """Check if a site is a branching point.

    The branch point definition is key for splitting MOFs into linker and nodes.
    Branch points are here defined as points
    that have at least three connections that do not lead to a tree or
    leaf node.

    Args:
        mof: MOF object
        index (int): index of site that is to be probed
        allow_metal (bool): If True it does not perform
            this check for metals (and just return False). Defaults to False.

    Returns:
        bool: True if this is a branching index
    """
    connected_sites = mof.get_neighbor_indices(index)

    if len(connected_sites) < 3:
        return False

    if (not allow_metal) and (index in mof.metal_indices):
        return False

    # lets store all the info in a numpy array
    sites = []
    for connected_site in connected_sites:
        leads_to_terminal = mof._leads_to_terminal((index, connected_site))
        is_terminal = mof._is_terminal(connected_site)
        in_metal_indices = connected_site in mof.metal_indices
        non_bridge_metal = has_non_bridge_path_with_metal(mof, index)
        sites.append([leads_to_terminal, is_terminal, in_metal_indices, non_bridge_metal])
    sites = np.array(sites).astype(bool)
    non_terminal_metal_connections = np.sum(~sites[:, 0] & ~sites[:, 1] & sites[:, 3])
    non_terminal_non_metal_connections = np.sum(~sites[:, 0] & ~sites[:, 1] & ~sites[:, 2])
    terminal_metal_connections = np.sum((sites[:, 0] | sites[:, 1]) & sites[:, 2])

    if terminal_metal_connections > 0:
        return False
    if (non_terminal_non_metal_connections >= 2) & (non_terminal_metal_connections > 0):
        return True

    return False


def filter_branch_points(
    mof: "mof.MOF", branching_indices  # noqa: F821 - forward reference
) -> List[int]:
    """Return a list of all branching points in the MOF."""
    # now, check if there are connected branching points
    # we need to clean that up as it does not make sense for splitting and the clustering
    # algorithm will not work (i believe).
    graph_ = mof.structure_graph.__copy__()
    graph_.structure = Structure.from_sites(graph_.structure.sites)
    to_delete = _not_relevant_structure_indices(mof.structure, branching_indices)
    graph_.remove_nodes(to_delete)
    mols, graphs, idx, centers, coordinates = get_subgraphs_as_molecules(
        graph_, return_unique=False
    )

    verified_indices = []

    for _, graph, index in zip(mols, graphs, idx):
        if len(index) == 1:
            verified_indices.extend(index)
        else:
            (
                good_by_distance,
                not_resolvable_by_distance,
                not_resolvable_graphs,
            ) = remove_not_closest_neighbor(mof, graph.graph, index)
            verified_indices.extend(good_by_distance)
            for not_resolvable_graph, not_resolvable_by_distance_index in zip(
                not_resolvable_graphs, not_resolvable_by_distance
            ):
                verified_indices.extend(
                    cluster_nodes(not_resolvable_graph, not_resolvable_by_distance_index, mof)
                )

    return verified_indices


def remove_not_closest_neighbor(
    mof: "mof.MOF", graph: nx.Graph, index: int  # noqa: F821 - forward reference
) -> List[int]:
    """Remove all nodes that are not the closest neighbor to the given index."""
    good_indices = []
    not_resolvable = []
    not_resolvable_graphs = []
    # divide indices into two groups, one with the closest neighbor metal and one with the rest
    minimum_distance, not_minimum_distance = rank_by_metal_distance(index, mof)
    # delete all nodes that are not the closest neighbor

    graph.remove_nodes_from(not_minimum_distance)
    # now, get new connected components
    for g in nx.connected_components(graph.to_undirected()):
        if len(g) == 1:
            good_indices.append(index[list(g)[0]])
        else:
            not_resolvable.append(itemgetter(*list(g))(index))
            not_resolvable_graphs.append(graph.subgraph(g))
    if len(good_indices) == 0:
        logger.warning("Could not find unique branching point")
    return good_indices, not_resolvable, not_resolvable_graphs


def has_bond_to_metal(mof: "MOF", index: int) -> bool:  # noqa: F821 - forward reference
    """Return True if the site has a bond to a metal."""
    return any(i in mof.metal_indices for i in mof.get_neighbor_indices(index))


def rank_by_metal_distance(idx, mof) -> Tuple[np.ndarray, np.ndarray]:
    """Rank the indices by the distance to the metal."""
    idx = np.array(idx)
    # ToDo: should use the path length
    # in this case, it simply means if there is one
    # direct single bond
    distances = -np.array([has_bond_to_metal(mof, i) for i in idx]).astype(int)

    minimum_distance = np.min(distances)
    return np.where(distances == minimum_distance)[0], np.where(distances != minimum_distance)[0]


def cluster_nodes(
    graph: nx.Graph, original_indices: List[int], mof: "mof.MOF"  # noqa: F821 - forward reference
) -> List[int]:
    g = nx.Graph(graph).to_undirected()
    terminal_nodes = []
    for node in g.nodes:
        if g.degree[node] == 1:
            terminal_nodes.append(node)
        if g.degree[node] > 2:
            raise ValueError("Unsupported connectivity of branching points.")

    if len(terminal_nodes) > 2:
        raise ValueError("Unsupported connectivity of branching points.")

    path = nx.shortest_path(g, terminal_nodes[0], terminal_nodes[1])
    # if odd return the middle node
    if len(path) % 2 == 1:
        return [original_indices[path[int(len(path) / 2)]]]

    return original_indices


def get_branch_points(mof: "mof.MOF") -> List[int]:  # noqa: F821 - forward reference
    """Get all branching points in the MOF.

    Args:
        mof (MOF): MOF object.

    Returns:
        List[int]: List of indices of branching points.
    """
    branch_points = [i for i in range(len(mof.structure)) if _is_branch_point(mof, i)]
    return filter_branch_points(mof, branch_points)



# -*- coding: utf-8 -*-
"""Create Python containers for node building blocks.

Here we understand metal clusters as nodes.
"""
from collections import defaultdict
from typing import Set

import numpy as np
from pymatgen.analysis.graphs import MoleculeGraph
from pymatgen.core import Molecule, Structure
from structuregraph_helpers.create import get_nx_graph_from_edge_tuples

from moffragmentor.fragmentor.molfromgraph import wrap_molecule

from .sbu import SBU
from ..fragmentor.splitter import _sites_and_classified_indices_from_indices
from ..utils import (
    _not_relevant_structure_indices,
    _reindex_list_of_tuple,
    get_neighbors_from_nx_graph,
)

__all__ = ["Node"]


def node_from_mof_and_indices(  # pylint:disable=too-many-locals, too-many-arguments
    cls, mof, node_indices, branching_indices, binding_indices, connecting_paths
):
    """Create a node from a MOF and a list of indices of different types."""
    node_indices = node_indices | connecting_paths  # This should actually not be necessary
    graph_ = mof.structure_graph.__copy__()
    graph_.structure = Structure.from_sites(graph_.structure.sites)
    to_delete = _not_relevant_structure_indices(mof.structure, node_indices)
    graph_.remove_nodes(to_delete)

    sites_and_indices = _sites_and_classified_indices_from_indices(mof, node_indices)
    relevant_indices = node_indices - set(sites_and_indices.hidden_vertices)

    selected_positions, selected_species, selected_edge_indices = [], [], []

    # give us a quick and easy way to filter the edges we want to keep
    # for this reason we will use both edge "partners" as keys
    # and have a list of indices in the original edges list as the value
    edge_dict = defaultdict(list)
    for i, egde in enumerate(sites_and_indices.edges):
        edge_partner_1, edge_partner_2 = egde
        edge_dict[edge_partner_1].append(i)
        edge_dict[edge_partner_2].append(i)

    for i, index in enumerate(node_indices):
        if index in relevant_indices:
            selected_positions.append(sites_and_indices.cartesian_coordinates[i])
            selected_species.append(sites_and_indices.species[i])
            edge_idxs = edge_dict[index]
            for edge_idx in edge_idxs:
                edge = sites_and_indices.edges[edge_idx]
                if (edge[0] in relevant_indices) and (edge[1] in relevant_indices):
                    selected_edge_indices.append(edge_idx)

    index_map = dict(zip(relevant_indices, range(len(relevant_indices))))
    molecule = Molecule(selected_species, selected_positions)
    selected_edges = [sites_and_indices.edges[i] for i in selected_edge_indices]
    selected_edges = _reindex_list_of_tuple(selected_edges, index_map)
    edge_dict = dict(zip(selected_edges, [None] * len(selected_edges)))
    molecule_graph = MoleculeGraph.with_edges(molecule, edge_dict)
    center = np.mean(molecule.cart_coords, axis=0)
    graph_branching_indices = branching_indices & node_indices
    closest_branching_index_in_molecule = []

    graph = get_nx_graph_from_edge_tuples(sites_and_indices.edges)
    for branching_index in branching_indices:
        if branching_index not in relevant_indices:

            # now we need to find the closest neighbor in the
            # set of vertices that are in the molecule
            # ToDo: this is currently an assumption that it is terminal and the
            # next partner then already is in the molecule,
            # we could recursively call or get all the paths and then get the shortest
            new_candidates = get_neighbors_from_nx_graph(graph, branching_index)[0]
            closest_branching_index_in_molecule.append(new_candidates)
        else:
            closest_branching_index_in_molecule.append(branching_index)
    idx = [i for i in relevant_indices if i in node_indices]

    mol_w_hidden, mapping = wrap_molecule(idx + sites_and_indices.hidden_vertices, mof)

    node = cls(
        molecule=mol_w_hidden,
        molecule_graph=molecule_graph,
        center=center,
        graph_branching_indices=graph_branching_indices,
        closest_branching_index_in_molecule=closest_branching_index_in_molecule,
        binding_indices=binding_indices,
        original_indices=idx + sites_and_indices.hidden_vertices,
        persistent_non_metal_bridged=sites_and_indices.persistent_non_metal_bridged_components,
        terminal_in_mol_not_terminal_in_struct=sites_and_indices.hidden_vertices,
        connecting_paths=connecting_paths,
        molecule_original_indices_mapping=mapping,
    )

    return node


class Node(SBU):
    """Container for metal cluster building blocks.

    Will typically automatically be constructured by the fragmentor.
    """

    @classmethod
    def from_mof_and_indices(  # pylint:disable=too-many-arguments
        cls,
        mof,
        node_indices: Set[int],
        branching_indices: Set[int],
        binding_indices: Set[int],
        connecting_paths: Set[int],
    ):
        """Build a node object from a MOF and some intermediate outputs of the fragmentation.

        Args:
            mof (MOF): The MOF to build the node from.
            node_indices (Set[int]): The indices of the nodes in the MOF.
            branching_indices (Set[int]): The indices of the branching points in the MOF
                that belong to this node.
            binding_indices (Set[int]): The indices of the binding points in the MOF
                that belong to this node.
            connecting_paths (Set[int]): The indices of the connecting paths in the MOF
                that belong to this node.

        Returns:
            A node object.
        """
        return node_from_mof_and_indices(
            cls,
            mof,
            node_indices,
            branching_indices,
            binding_indices,
            connecting_paths=connecting_paths,
        )

    def __repr__(self) -> str:
        """Return a string representation of the node."""
        return f"Node ({self.composition})"



# -*- coding: utf-8 -*-
"""Test functions related to running systre"""
from moffragmentor.utils.systre import _parse_systre_lines, run_systre


def test_systre_call(get_cgd_file):
    """Test the call to systre and the parsing of the output"""
    input_file = get_cgd_file
    systre_output = run_systre(input_file)
    assert systre_output["space_group"] == "Fm-3m"
    assert systre_output["rcsr_code"] == "tbo"
    assert systre_output["relaxed_cell"] == [4.89898, 4.89898, 4.89898]
    assert len(systre_output["relaxed_node_positions"]) == 2
    assert isinstance(systre_output["relaxed_node_positions"], dict)


def test_systre_parser(get_systre_output, get_systre_output2):
    systre_output = _parse_systre_lines(get_systre_output.split("\n"))
    assert systre_output["space_group"] == "Fm-3m"
    assert systre_output["rcsr_code"] == "tbo"
    assert systre_output["relaxed_cell"] == [4.89898, 4.89898, 4.89898]
    assert len(systre_output["relaxed_node_positions"]) == 2
    assert isinstance(systre_output["relaxed_node_positions"], dict)

    systre_output2 = _parse_systre_lines(get_systre_output2.split("\n"))
    assert systre_output2["space_group"] == "p4mm"
    assert systre_output2["rcsr_code"] == "mtf"
    assert systre_output2["relaxed_cell"] == [3.05792, 3.05792, 90.0000]



# -*- coding: utf-8 -*-
"""Compute dimensionality of MOF building blocks.

One important piece of information about the SBUs is their dimenaionality,
e.g., if they are clusters, if they form rods, or if they form sheets.
We use the algorithm proposed by Larsen et al.,
see https://journals.aps.org/prmaterials/pdf/10.1103/PhysRevMaterials.3.034003
"""
from typing import Iterable

from pymatgen.analysis.dimensionality import get_dimensionality_larsen
from pymatgen.analysis.graphs import StructureGraph

from ..utils import remove_all_nodes_not_in_indices


def get_structure_graph_dimensionality(structure_graph: StructureGraph) -> int:
    """Use Larsen's algorithm to compute the dimensionality"""
    return get_dimensionality_larsen(structure_graph)


def get_sbu_dimensionality(mof: "MOF", indices: Iterable[int]) -> int:  # noqa:F821
    """Compute the dimensionality of an SBU.

    SBU ischaracterized as a subset of indices of a MOF

    Args:
        mof (MOF): moffragmentor MOF instance.
        indices (Iterable[int]): indices of the SBU.

    Returns:
        int: dimensionality of the SBU.
    """
    structure_graph = mof.structure_graph.__copy__()
    remove_all_nodes_not_in_indices(structure_graph, indices)
    return get_structure_graph_dimensionality(structure_graph)



# -*- coding: utf-8 -*-
"""Extraction of pymatgen Molecules from a structure for which we know the branching points."""
from collections import namedtuple

import networkx as nx
import numpy as np
from pymatgen.analysis.graphs import MoleculeGraph
from pymatgen.core import Molecule

from .solventlocator import _locate_bound_solvent
from ..utils import (
    _get_cartesian_coords,
    _get_molecule_edge_label,
    _get_vertices_of_smaller_component_upon_edge_break,
    _reindex_list_of_tuple,
    revert_dict,
)

_SitesAndIndicesOutput = namedtuple(
    "SitesAndIndicesOutput",
    [
        "cartesian_coordinates",
        "species",
        "edges",
        "index_mapping",
        "hidden_vertices",
        "persistent_non_metal_bridged_components",
    ],
)


def _sites_and_classified_indices_from_indices(mof, indices: set) -> _SitesAndIndicesOutput:
    """Extract the sites and indices of a molecule from a set of indices.

    Given the indices that we identified to belong to the group
    1) Build a molecule from the structure, handling the periodic images
    2) Build the molecule graph from the structure graph,
        handling the periodic images
    3) Flag vertices in the graph that we call "hidden". Those have only one neighbor
        in the molecule but multiple in the structure.
    This happens, for example, in MOF-74 where the "connecting site"
        is not on the carboxy carbon
        given that one of the carboxy Os is not connected to the metal cluster
    To get reasonable nodes/linkers we will remove these from the node.
    We also flag indices that are connected via a bridge that
        is also a bridge in the original graph.
        This is, for example, the case for one carboxy O in MOF-74.

    Args:
        mof (MOF): instance of a MOF objec
        indices (set): set of indices that we identified to belong to the group

    Returns:
        _SitesAndIndicesOutput: namedtuple containing the following fields:
            cartesian_coordinates (np.array): cartesian coordinates of the sites
            species (list): species of the sites
            edges (list): edges of the molecule
            index_mapping (dict): mapping from indices in the structure to indices in the molecule
            hidden_vertices (set): indices of vertices in the molecule that should be hidden
                (num_neighbors_in_molecule == 1) and len(neighbors) > 1)
            persistent_non_metal_bridged_components (set): indices of vertices in the molecule that are
    """
    new_positions = []
    persistent_non_metal_bridged_components = []
    hidden_vertices = []
    # we store edges as list of tuples as we can
    # do a cleaner loop for reindexing
    edges = []
    index_mapping = {}

    # cast set to list, so we can loop
    indices = list(indices)
    species = []
    for new_idx, idx in enumerate(indices):
        new_positions.append(_get_cartesian_coords(mof, indices[0], idx))
        index_mapping[idx] = new_idx
        # This will fail for alloys/fractional occupancy
        # ToDo: check early if this might be an issue
        species.append(str(mof.structure[idx].specie))
        neighbors = mof.structure_graph.get_connected_sites(idx)
        # counter for the number of neighbors that are also part
        # of the molecule, i.e., in `indices`
        num_neighbors_in_molecule = 0
        for neighbor in neighbors:
            # we also identified the neighbor to be part of the molecule
            if neighbor.index in indices:
                edges.append(_get_molecule_edge_label(idx, neighbor.index))
                num_neighbors_in_molecule += 1
        if (num_neighbors_in_molecule == 1) and len(neighbors) > 1:
            hidden_vertices.append(idx)

    reindexed_edges = _reindex_list_of_tuple(edges, index_mapping)
    edge_dict = dict(zip(reindexed_edges, [None] * len(reindexed_edges)))

    mol = Molecule(species, new_positions)
    graph = MoleculeGraph.with_edges(mol, edge_dict)

    new_bridges = set(nx.bridges(nx.Graph(graph.graph.to_undirected())))
    new_index_to_old_index = revert_dict(index_mapping)

    solvent = _locate_bound_solvent(mof, indices)
    any_bound_solvent_index = set(sum(solvent["solvent_indices"], []))
    # Here we now need to get the bound solvent components
    # to make sure we can exlude them
    # ToDo: we can then export this result and skip the calculation
    # in the fragment function

    for bridge in new_bridges:
        component = _get_vertices_of_smaller_component_upon_edge_break(
            graph.graph.to_undirected(), bridge
        )
        indices = [new_index_to_old_index[i] for i in component]
        if not set(indices) & any_bound_solvent_index:
            persistent_non_metal_bridged_components.append(indices)

    persistent_non_metal_bridged_components_old_idx = []
    for subset in persistent_non_metal_bridged_components:
        persistent_non_metal_bridged_components_old_idx.append(subset)

    return _SitesAndIndicesOutput(
        np.array(new_positions),
        species,
        edges,
        index_mapping,
        hidden_vertices,
        persistent_non_metal_bridged_components_old_idx,
    )
